import json
import time
import random as r
import pokeapi
from prediction import Prediction
from trade import Trade
import discord
from battle import Battle
import asyncio

TRADE_TIME = 15
BATTLE_REQUEST_TIME = 60
STARTING_COIN = 200
COIN_GAIN = 75
COIN_VARIABILITY = 25
COIN_TIP = 25
COLLECT_TIME = 79200
TIP_TIME = 79200
FREEROLL_TIME = 79200
POKE_COST = 50
OMEGALUL = "<:omega:844632595110166528>"

_current_battle = None
_trade_offer = None
_prediction = None
_db = {}


# helpers
def _getTime():
    return int(time.time())


def _formatRemainingTime(remainingTime):
    minutes, seconds = divmod(remainingTime, 60)
    hours, minutes = divmod(minutes, 60)

    timeUnits = []
    if hours:
        timeUnits.append(f"{hours} hour{'s' if hours > 1 else ''}")
    if minutes:
        timeUnits.append(f"{minutes} minute{'s' if minutes > 1 else ''}")
    if seconds:
        timeUnits.append(f"{seconds} second{'s' if seconds > 1 else ''}")

    if len(timeUnits) == 0:
        return "0 seconds"
    elif len(timeUnits) == 1:
        return timeUnits[0]
    elif len(timeUnits) == 2:
        return f"{timeUnits[0]} and {timeUnits[1]}"
    return f"{timeUnits[0]}, {timeUnits[1]}, and {timeUnits[2]}"


def _getID(user):
    return str(user.id)


def _hasPokemon(userid, pokeName):
    id = pokeapi.getID(pokeName)
    if not id:
        return False
    return id in _db[userid]["inventory"]


def _cleanString(s):
    return "".join(c.lower() for c in s if not c.isspace() and c != "_")


def _getSuggestionText(name):
    matches = [f"**{n}**" for n in pokeapi.getSuggestions(name)]
    if not matches:
        return "Try again, I have no idea what pokemon you meant."
    matchtext = ", or ".join(matches)
    return f"Did you mean {matchtext}?"


def _coinTable():
    next = 1
    lst = []
    for key, val in _db.items():
        lst.append((val["name"], val["coin"]))
    lst.sort(reverse=True, key=lambda x: x[1])
    stringer = " \t\t\t\t\t  nickcoin\n"
    stringer += "-------------------------------------------------------\n"

    for person in lst:
        nextpart = "|  {:^2s} |  {:^24s} |  {:^16s} |".format(
            str(next), str(person[0]), str(person[1])
        )
        stringer += nextpart + "\n"
        next += 1
    stringer += "-------------------------------------------------------\n"
    return "```" + stringer + "```"


def _getCommandContent(contents, hasAmount=False):
    if not hasAmount:
        return " ".join(contents[1:])
    return " ".join(contents[:-1][1:])


# commands
async def _start(send, contents, user):
    userid = _getID(user)
    if userid in _db:
        await send("You already have an account.")
        return False
    _db[userid] = {
        "coin": STARTING_COIN,
        "collecttime": 0,
        "name": user.name,
        "tiptime": 0,
        "freerolltime": 0,
        "inventory": {},
        "flex": {},
    }
    await send("Hi, welcome to Nickcoin. Run {} to see your balance.", "!coin")
    return True


async def _coin(send, contents, user):
    await send(_coinTable())
    return False


async def _collect(send, contents, user):
    userid = _getID(user)
    oldtime = int(_db[userid]["collecttime"])
    remaining = _getTime() - oldtime
    if remaining < COLLECT_TIME:
        timeLeft = _formatRemainingTime(COLLECT_TIME - remaining)
        await send("You have to wait {} before collecting.", timeLeft)
        return False
    amnt = COIN_GAIN + r.randint(-COIN_VARIABILITY, COIN_VARIABILITY)
    _db[userid]["coin"] += amnt
    _db[userid]["collecttime"] = _getTime()
    await send(
        "Collected {} nickcoins! You now have {} nickcoins.", amnt, _db[userid]["coin"]
    )
    return True


async def _tip(send, contents, user):
    userid = _getID(user)
    if len(contents) <= 1:
        await send("Invalid command")
        return
    oldtime = int(_db[userid]["tiptime"])
    remaining = _getTime() - oldtime
    if remaining < TIP_TIME:
        time_left = _formatRemainingTime(TIP_TIME - remaining)
        await send("You have to wait {} before tipping.", time_left)
        return False
    name = _getCommandContent(contents)
    for key, value in _db.items():
        cleaned_name = _cleanString(value["name"])
        # you cannot tip yourself
        if cleaned_name == _cleanString(name) and cleaned_name != _cleanString(
            user.name
        ):
            _db[key]["coin"] += COIN_TIP
            await send("tipped {} to {}!", COIN_TIP, name)
            _db[userid]["tiptime"] = _getTime()
            await send(_coinTable())
            return True
    await send("Couldn't find that user.")
    return False


async def _give(send, contents, user):
    if len(contents) < 3 or not contents[-1].isnumeric():
        await send("Invalid command. Should be !give USERNAME AMOUNT")
        return False
    name = _getCommandContent(contents, hasAmount=True)
    amnt = contents[-1]
    for key, value in _db.items():
        if _cleanString(value["name"]) == _cleanString(name):
            _db[key]["coin"] += int(amnt)
            await send("gave {} {} nickcoin", name, amnt)
            return True
    await send("couldn't find that user.")
    return False


async def _pay(send, contents, user):
    userid = _getID(user)
    if len(contents) < 3 or not contents[-1].isnumeric():
        await send("Invalid command. Should be !pay USERNAME AMOUNT")
        return False
    name = _getCommandContent(contents, hasAmount=True)
    amnt = contents[-1]
    if _db[userid]["coin"] < int(amnt):
        await send("You cant afford to give that much.")
        return False
    for key, value in _db.items():
        if _cleanString(value["name"]) == _cleanString(name):
            _db[key]["coin"] += int(amnt)
            _db[userid]["coin"] -= int(amnt)
            await send("{} paid {} {} nickcoin", user.name, name, amnt)
            return True
    await send("couldn't find that user.")
    return False


# pokemon
async def __executeRoll(send, user, gen):
    userid = _getID(user)
    pokeid = [pokeapi.roll(gen)]
    _putInInv(_db[userid]["inventory"], pokeid[0])
    pic = discord.File(pokeapi.getImages(pokeid), "_.png")
    time.sleep(3)
    await send("{} rolled {}", user.name, pokeapi.getName(pokeid[0]), file=pic)


def _putInInv(inv, pokeid):
    if pokeid in inv:
        inv[pokeid] += 1
    else:
        inv[pokeid] = 1


def _removeFromInv(inv, pokeid):
    inv[pokeid] -= 1
    if inv[pokeid] == 0:
        inv.pop(pokeid)


def _verifyFlex(inv, flex):
    for pokeid, occur in flex.items():
        if occur > inv[pokeid]:
            return False
    return True


def _rollFn(gen):
    async def _roll(send, contents, user):
        userid = _getID(user)
        if _db[userid]["coin"] < POKE_COST:
            await send("Too Poor. It costs {} nickcoin for a roll.", POKE_COST)
            return False
        await __executeRoll(send, user, gen)
        _db[userid]["coin"] -= POKE_COST
        return True

    return _roll


async def _freeroll(send, contents, user):
    userid = _getID(user)
    oldtime = int(_db[userid]["freerolltime"])
    remaining = _getTime() - oldtime
    if remaining < FREEROLL_TIME:
        timeLeft = _formatRemainingTime(FREEROLL_TIME - remaining)
        await send("You have to wait {} before getting a free roll.", timeLeft)
        return False
    await __executeRoll(send, user, 0)
    _db[userid]["freerolltime"] = _getTime()
    return True


def _invToList(inv):
    inv_list = []
    for pokeid, occur in inv.items():
        inv_list.extend([pokeid] * occur)
    return inv_list


async def _inv(send, contents, user):
    name = user.name
    inv = []
    if len(contents) > 1:
        name = _getCommandContent(contents)
    for key, value in _db.items():
        if _cleanString(value["name"]) == _cleanString(name):
            inv = _invToList(value["inventory"])  # building a list of pokeids
            if not inv:  # empty list
                await send("You have no pokemon. Sad")
                return False

            pic = discord.File(pokeapi.getImages(inv), "_.png")
            await send("{}'s inventory:", name, file=pic)
            return False
    await send("Couldn't find that user.")
    return False


async def _evolve(send, contents, user):
    userid = _getID(user)
    if len(contents) != 2:
        await send("Invalid command")
        return False
    name = _getCommandContent(contents)
    id = pokeapi.getID(name)
    if not id:
        await send(_getSuggestionText(name))
        return False
    inv = _db[userid]["inventory"]
    if not inv:
        await send("You don't have any pokemon.")
        return False
    evoid = pokeapi.evolve(id)
    if evoid == "-1":
        await send("That pokemon cannot evolve.")
        return False
    amntToEvolve = 3
    if name == "magikarp" or name == "feebas":
        amntToEvolve = 10
    if id not in inv or inv[id] < amntToEvolve:
        await send("You don't have enough.")
        return False
    for _ in range(amntToEvolve):
        _removeFromInv(inv, id)
    newevo = pokeapi.getName(evoid)
    _putInInv(inv, evoid)
    pic = discord.File(pokeapi.getImages([evoid]), "_.png")
    await send("Your {} evolved into a {}!", name, newevo, file=pic)
    return True


async def _trade(send, contents, user):
    global _trade_offer
    if len(contents) < 3:
        await send("not valid. should be !trade GIVEAWAY RECEIVE")
        return False
    poke1, poke2 = contents[1], contents[2]
    if _trade_offer:
        if _getTime() - _trade_offer.time > TRADE_TIME:
            _trade_offer = None
            await send("trade expired")
            return False
        await send(
            "{} is offering to give their {} to obtain a {}. {} to accept",
            user.name,
            poke1,
            poke2,
            "!accept",
        )
        return False
    if not pokeapi.getID(poke2):
        await send(_getSuggestionText(poke2))
        return False

    if not _hasPokemon(_getID(user), poke1):
        await send("You dont have {}.", poke1)
        return False
    _trade_offer = Trade(user, poke1, poke2, _getTime())
    await send(
        "{} is offering to give their {} to obtain a {}. {} to accept",
        user.name,
        poke1,
        poke2,
        "!accept",
    )
    return False


async def _accept(send, contents, user):
    global _trade_offer
    if not _trade_offer:
        await send("no trades open.")
        return False
    if _getTime() - _trade_offer.time > TRADE_TIME:
        _trade_offer = None
        await send("trade expired")
        return False
    if _hasPokemon(_getID(user), _trade_offer.poke2):
        pokeID1 = pokeapi.getID(_trade_offer.poke1)
        pokeID2 = pokeapi.getID(_trade_offer.poke2)
        userid2 = _getID(user)
        userid1 = _getID(_trade_offer.user)
        _putInInv(_db[userid2]["inventory"], pokeID1)
        _removeFromInv(_db[userid2]["inventory"], pokeID2)
        _putInInv(_db[userid1]["inventory"], pokeID2)
        _removeFromInv(_db[userid1]["inventory"], pokeID1)
        await send(
            "{} traded {} for {} with {}!",
            _trade_offer.user.name,
            _trade_offer.poke1,
            _trade_offer.poke2,
            user.name,
        )
        _trade_offer = None
        return True
    await send("You can't do this trade, {}.", user.name)
    return False


async def _flex(send, contents, user):
    name = user.name
    if len(contents) > 1:
        name = _getCommandContent(contents)
    for key, value in _db.items():
        if value["name"] == name:
            flex_inv = value["flex"]
            if not flex_inv:
                await send("{} haven't set up your flex. {}", value["name"], "!setflex")
                return False
            if not _verifyFlex(value["inventory"], flex_inv):  # problem in flex vs inv
                await send(
                    "{}'s flex is not up to date. {} to update.",
                    value["name"],
                    "!setflex",
                )
                return False
            flexlist = _invToList(flex_inv)
            pic = discord.File(pokeapi.getImages(flexlist), "_.png")
            await send("{}'s pokemon:", name, file=pic)
            return False
    await send("Couldn't find that user.")
    return False


async def _setflex(send, contents, user):
    userid = _getID(user)
    inv = _db[userid]["inventory"]
    ids = {}
    for poke in contents[1:]:
        id_to_add = pokeapi.getID(poke)
        if not id_to_add:
            await send(_getSuggestionText(poke))
            return False
        if id_to_add not in inv:
            await send("nice try.")
            return False

        _putInInv(ids, id_to_add)
        if ids[id_to_add] > inv[id_to_add]:
            await send("nice try.")
            return False
    userid = _getID(user)
    _db[userid]["flex"] = ids
    await send("Flex updated!")
    return True


# predictions
async def _battle(send, contents, user):
    global _current_battle
    # checks that the command was given properly
    if len(contents) == 1:
        decl_max = "any number of"
    elif len(contents) == 2 and contents[1].isnumeric():
        decl_max = contents[1]
    else:
        await send(
            "Invalid command. Format should be {}", "!battle [max declare number]"
        )
        return False
    # checks if there is a current battle ongoing already or if battle request expired
    if _current_battle:
        if _getTime() - _current_battle.time > BATTLE_REQUEST_TIME:
            _current_battle = None
            await send("Battle request expired")
            return False
        await send("A battle is already ongoing!")
        return False
    else:  # starts a instance of the Battle Class for this new battle
        await send(
            (
                "{} wants to engage in a pokemon battle. "
                "Who dares to accept? {} to accept."
            ),
            user.name,
            "!baccept",
        )
        _current_battle = Battle(user.name, _getTime(), decl_max)
        return False


async def _accept_battle(send, contents, user):
    global _current_battle
    if not _current_battle:  # checks if there are current battle requests or not
        await send("No current battle requests.")
        return False
        # checks if the battle request expired
    if _getTime() - _current_battle.time > BATTLE_REQUEST_TIME:
        _current_battle = None
        await send("Battle request expired")
        return False
    # checks to make sure that there are no current battles already ongoing
    if _current_battle.is_happening:
        await send("There is already a battle ongoing!")
        return False
    # make sure that p1 doesn't accept their own battle
    if _current_battle.getP1Name() == user.name:
        await send("You are not your worst enemy.")
        return False
    else:  # starts the battle
        _current_battle.acceptPlayerTwo(user.name)
        await send(
            (
                "A battle has begun between {} and {}. {} to declare {} "
                "pokemon. Hold on to your hats, it's going to be a wild one!"
            ),
            _current_battle.getP1Name(),
            _current_battle.getP2Name(),
            "!declare",
            _current_battle.decl_max,
        )
        _current_battle.is_happening = True
        return False


async def _declare(send, contents, user):
    global _current_battle
    if not _current_battle:
        await send("There isn't a battle going on!")
        return False
    poke_names = contents[1:]
    player = _current_battle.getPlayerByName(user.name)
    userid = _getID(user)
    inv = _db[userid]["inventory"]
    if not poke_names:
        await send("You forgot to declare a pokemon.")
        return False
    if not player:
        await send("{}, you are not a player in this battle. Bye", user.name)
        return False
    for poke in poke_names:
        pokeid = pokeapi.getID(poke)
        if not pokeid:
            await send(_getSuggestionText(poke))
            return False

        if pokeid not in inv or poke_names.count(poke) > inv[pokeid]:
            await send(
                "Nice try. You don't have enough {}. {} to try again.", poke, "!declare"
            )
            return False
    decl_max = _current_battle.decl_max
    if decl_max.isnumeric() and len(poke_names) > int(decl_max):
        await send("You can only declare {} pokemon!", decl_max)
        return False

    re = ""
    if player.hasDeclared():  # No pokes declared yet
        re = "re-"
    player.declare_pokes(poke_names)
    await send("{} has " + f"{re}declared their pokemon!", user.name)
    return False


async def _start_battle(send, contents, user):
    async def sendFn(msg, img_list=[], delete_after=None):
        pic = None
        if len(img_list) > 0:
            pic = discord.File(
                pokeapi.getBattlePokemonImages(
                    [pokeapi.nameToID[poke_name] for poke_name in img_list]
                ),
                "_.png",
            )
        sent_message = await send(msg, file=pic, delete_after=delete_after)
        return sent_message

    # check that p1, p2, p1pokes, p2pokes all exist and starts the battle.
    global _current_battle
    if _current_battle and _current_battle.is_happening:
        await _current_battle.compute_and_send_script(sendFn)
        _current_battle = None
    else:
        await send(
            "Silly {}, " + f"there's n{OMEGALUL} battle happening right now!",
            user.name,
        )


async def _end_battle(send, contents, user):
    global _current_battle
    if _current_battle.getPlayerByName(user.name):
        _current_battle.is_happening = False
        await asyncio.sleep(6)
        _current_battle = None
    else:
        await send("There is a time and place for everything but not now!")


async def _who_has(send, contents, user):
    names = contents[1:]

    if len(contents) < 2:
        await send("Invalid command")
        return False

    # short_inv = {"user name" : ["2 dratini", "3 charmander"]}

    short_inv = {}
    for userid, value in _db.items():
        user_name = value["name"]
        short_inv[user_name] = []
        for poke in names:
            pokeid = pokeapi.getID(poke)
            if not pokeid:
                await send(_getSuggestionText(poke))
                return False
            if _hasPokemon(userid, poke):
                num_occur = value["inventory"][pokeid]
                short_inv[user_name].append(f"{num_occur} {poke}")

        # this user doesn't have all the pokemon requested by who_has
        if len(short_inv[user_name]) != len(names):
            short_inv.pop(user_name)
        else:
            short_inv[user_name] = " and ".join(short_inv[user_name])

    final_msg = ""
    final_args = []
    for user_name, sent in short_inv.items():
        final_msg += "{} has {}\n"
        final_args.extend([user_name, sent])
    if not final_msg:
        final_msg = "no one has {}!"
        final_args.append(" and ".join(names))
    await send(final_msg, *final_args)
    return False


# predictions
async def _start_prediction(send, contents, user):
    global _prediction
    if _prediction:
        await send(_prediction.state())
        return False
    if len(contents) < 2:
        await send("Please provide an event description.")
        return False
    eventStr = _getCommandContent(contents)
    _prediction = Prediction(eventStr, user, _db)
    await send(
        "PREDICTION FOR {} STARTED!\n" + "To join: !believe XXX or !doubt XXX",
        _prediction.event,
    )
    return False


async def _close(send, contents, user):
    if not _prediction:
        await send("No prediction right now.")
        return False
    await send(_prediction.close(user))
    await send(_coinTable())
    return True


def _placebetfn(side):
    async def placebet(send, contents, user):
        if not _prediction:
            await send("No prediction right now.")
            return False
        if len(contents) != 2:
            await send("Invalid command")
            return False
        amnt = _getCommandContent(contents)
        if not amnt.isdecimal():
            await send("Has to be a number")
            return False
        num = int(amnt)
        if num == 0:
            await send("Has to be nonzero.")
            return False
        if side == "b":
            await send(_prediction.believe(user, num))
        else:
            await send(_prediction.doubt(user, num))
        return False

    return placebet


def _payoutfn(side):
    async def payout(send, contents, user):
        global _prediction
        if not _prediction:
            await send("No prediction right now.")
            return False
        await send(_prediction.payout(user, side))
        if _prediction.finished:
            _prediction = None
            await send(_coinTable())
            return True
        return False

    return payout


_commandMap = {
    "!start": _start,
    "!coin": _coin,
    "!collect": _collect,
    "!roll": _rollFn(0),
    "!roll1": _rollFn(1),
    "!roll2": _rollFn(2),
    "!roll3": _rollFn(3),
    "!roll4": _rollFn(4),
    "!freeroll": _freeroll,
    "!tip": _tip,
    "!trade": _trade,
    "!accept": _accept,
    "!give": _give,
    "!pay": _pay,
    "!inv": _inv,
    "!evolve": _evolve,
    "!flex": _flex,
    "!setflex": _setflex,
    "!predict": _start_prediction,
    "!believe": _placebetfn("b"),
    "!doubt": _placebetfn("d"),
    "!close": _close,
    "!WINBELIEVE": _payoutfn("b"),
    "!WINDOUBT": _payoutfn("d"),
    "!battle": _battle,
    "!baccept": _accept_battle,
    "!declare": _declare,
    "!bstart": _start_battle,
    "!battle_end": _end_battle,
    "!whohas": _who_has,
}


# backend API
def load_db():
    global _db
    with open("save.txt", "r") as file:
        data = file.read()
        if data:
            _db = json.loads(data)


def user_exists(user):
    return _getID(user) in _db


def getCommands():
    return _commandMap


def persist():
    f = open("save.txt", "w")
    f.write(json.dumps(_db))
    f.close()
