import json
import random
from PIL import Image
import io
import csv
from difflib import get_close_matches

IMAGE_H = 90
IMAGE_W = 90
POKES_PER_ROW = 10


def _getPath(id):
    return dex[str(id)]["sprite"]


def _getRarity(id):
    return dex[str(id)]["rarity"]


def _getImage(id):
    return Image.open(_getPath(id)).convert("RGBA")


# API
def roll(gen):
    # use ids[x:y] and probs[x:y] to pick a range.
    # ids is list of integers for pokedex nums.
    # probs is list of decimals.
    # they are sorted correctly.
    rand_num = random.randrange(8192)
    shiny_add_amt = 0
    if rand_num == 69:  # 1/8192 chance
        shiny_add_amt = 1000
    ids, probs = _idsToProb[gen]
    # now it returns an integer inside a string
    return str(random.choices(ids, probs)[0] + shiny_add_amt)


def getImages(idcs):  # takes a list of strings
    numIms = len(idcs)
    if numIms == 0:
        return None
    output = Image.new(
        "RGBA",
        (
            IMAGE_W * min(numIms, POKES_PER_ROW),
            ((numIms + POKES_PER_ROW - 1) // POKES_PER_ROW) * IMAGE_H,
        ),
        (0, 0, 0, 0),
    )

    def imgSortPrio(x):
        return (_getRarity(x), getName(x))

    for count, idx in enumerate(sorted(idcs, key=imgSortPrio, reverse=True)):
        im = _getImage(idx)
        output.paste(
            im,
            (
                IMAGE_W * (count % min(numIms, POKES_PER_ROW)),
                (count // POKES_PER_ROW) * IMAGE_H,
            ),
        )
    b = io.BytesIO()
    output.save(b, "PNG")
    b.seek(0)
    return b


def evolve(idx):
    res = dex[str(idx)]["evolution"]
    rand = random.randint(0, len(res) - 1)
    return str(res[rand])  # returns an integer inside of string of the pokemon id


def getName(idx):  # takes in an integer
    if str(idx) not in dex:
        return ""
    return dex[str(idx)]["name"]  # returns a string with pokemon


def getID(name):
    if name not in nameToID:
        return ""
    return nameToID[name]  # returns an integer in a string -- '15'


def getSuggestions(name):
    return get_close_matches(
        name,
        nameToID.keys(),
    )


def getRarity(idx):
    if str(idx) not in dex:
        return ""
    return dex[str(idx)]["rarity"]


def getTypes(idx):
    if str(idx) not in dex:
        return ""
    return dex[str(idx)]["types"]


def getStats(idx):
    if str(idx) not in dex:
        return ""
    return dex[str(idx)]["stats"]


def getBattlePoke(idx):
    stats = {key: (value * 2 + 52) + 5 for key, value in getStats(idx).items()}
    stats["hp"] += 105
    return {
        "stats": stats,
        "rarity": getRarity(idx),
        "types": getTypes(idx),
    }


def getBattlePokemonImages(idcs):
    numIms = len(idcs)
    if numIms == 0:
        return None
    output = Image.new(
        "RGBA",
        (IMAGE_W * min(numIms, POKES_PER_ROW), (1 + numIms // POKES_PER_ROW) * IMAGE_H),
        (0, 0, 0, 0),
    )
    for count, idx in enumerate(idcs):
        im = _getImage(idx)
        output.paste(
            im,
            (
                IMAGE_W * (count % min(numIms, POKES_PER_ROW)),
                (count // POKES_PER_ROW) * IMAGE_H,
            ),
        )
    b = io.BytesIO()
    output.save(b, "PNG")
    b.seek(0)
    return b


# dex maps from (pokemon index) => pdict
# pdict is a dictionary with:
# name(str), sprite(str url), rarity (str int), evolution(list int)
dex = {}
rarities = {}
nameToID = {}  # keys are strings with names and values of strings of integers (ids)

with open("data/pokedex.json", "r") as file:
    data = file.read()
    if data:
        dex = json.loads(data)
    else:
        raise Exception("no data in pokedex.json")
with open("data/rarities.csv", "r") as file:
    csv_reader = csv.DictReader(file)
    line_count = 1
    for row in csv_reader:
        dex[row["number"]]["rarity"] = int(row["rarity"])
        line_count += 1
        if line_count == 494:
            break

_dex_keys, _probabilities = [], []
for poke_id, pokedict in dex.items():
    _dex_keys.append(int(poke_id))
    nameToID[pokedict["name"]] = poke_id
    if pokedict["rarity"] > 0:
        _probabilities.append(1.0 / pokedict["rarity"])
    else:
        _probabilities.append(0)

_ids, _probs = zip(*sorted(zip(_dex_keys, _probabilities)))
_idsToProb = {
    0: (_ids, _probs),
    1: (_ids[0:151], _probs[0:151]),
    2: (_ids[151:251], _probs[151:251]),
    3: (_ids[251:386], _probs[251:386]),
    4: (_ids[386:494], _probs[386:494]),
}
